from PyQt5 import QtGui, QtCore, QtWidgets
import pyqtgraph as pg
from pyqtgraph import Point
import numpy as np
from skimage.morphology import disk, dilation, erosion
from skimage.draw import line_aa
from PIL import Image
from PIL import ImageDraw as ID
import math



class ImageDraw(pg.ImageItem):
    sigImageChanged = QtCore.pyqtSignal()

    def __init__(self, image = None, viewbox = None, parent = None, lut = None, **kargs):
        super(ImageDraw, self).__init__()
        self.levels = np.array([0,255])
        self.lut = lut
        self.axisOrder = "row-major"
        self.removable = False

        self.colorNum = None

        self.parent = parent
        self.setDrawKernel(kernel_size = self.parent.brush_size)
        self.parent.strokes = []
        self.parent.in_addregion_stroke = False
        self.brushColorSelected = False
        self.parent.currStroke = []
        
        self.strokeAppended = True
    
    def turnOnDraw(self):
        print("drawing on")
        self.setDefaultColorNum()
    
    def turnOffDraw(self):
        self.colorNum = None
        self.brushColorSelected = False
    
    def setDefaultColorNum(self):
        if self.parent.drawType == "Eraser":
            self.colorNum = 0
        else:
            self.colorNum = self.parent.currMask.max() + 1

    def setDrawKernel(self, kernel_size=3):
        bs = kernel_size
        kernel = np.ones((bs,bs), np.uint8)
        self.drawKernel = kernel
        self.drawKernelCenter = [int(np.floor(kernel.shape[0]/2)),
                                 int(np.floor(kernel.shape[1]/2))]
    
    def mouseClickEvent(self, ev):
        print("click event")
        if not self.parent.imLoaded:
            return
        
        cellNum = self.parent.currMask[int(ev.pos().y()), int(ev.pos().x())]
        
        if ev.button() == QtCore.Qt.LeftButton and self.parent.maskOn:
            if cellNum!=0:
                self.parent.selectCell(ev.pos())
        
        # mask must be on for drawing
        elif ev.button() == QtCore.Qt.RightButton and not self.parent.probOn and self.parent.maskOn:
            
            if self.parent.drawType != "":
                
                # different brush modes
                if self.parent.drawType == "Brush":
                    self.colorNum = self.parent.selectedCells[0] if self.parent.selectedCells else self.parent.currMask.max()+1
                    self.drawAt(ev.pos())
                
                elif self.parent.drawType == "Outline":
                    if not self.parent.in_addregion_stroke:
                        self.colorNum = self.parent.selectedCells[0] if self.parent.selectedCells and self.parent.selectedCells[0] != 0 else self.parent.currMask.max()+1
                        ev.accept()
                        self.create_start(ev.pos())
                        self.parent.in_addregion_stroke = True
                        self.drawAt(ev.pos(), ev)
                        self.strokeAppended = False
                    else:
                        print("outline click event second if")
                        ev.accept()
                        self.end_stroke()
                        self.parent.in_addregion_stroke = False

                elif self.parent.drawType == "Eraser" and not self.brushColorSelected:
                    self.colorNum = 0
                    self.drawAt(ev.pos())
                elif self.brushColorSelected:
                    self.drawAt(ev.pos())
                
                self.brushColorSelected = True

    def mouseDragEvent(self, ev):
        print("drag")
        if ev.button() == QtCore.Qt.LeftButton:
            self.parent.view.mouseDragEvent(ev)
            ev.ignore()
            return
        
        if not self.parent.imLoaded:
            return
        
        # mask must be on for drawing
        if ev.button() ==  QtCore.Qt.RightButton and (self.parent.drawType == "Brush" or self.parent.drawType == "Eraser"):
            ev.accept()
            if self.parent.drawType == "Eraser":
                self.colorNum = 0
            else:
                self.colorNum = self.parent.selectedCells[0] if self.parent.selectedCells else self.colorNum

            if not self.parent.probOn and self.parent.maskOn:
                if ev.isStart():
                    self.drawAt(ev.pos())
                elif ev.isFinish():
                    return
                else:
                    self.paintBrush(ev.lastPos(),ev.pos())
                return
    
    def hoverEvent(self, ev):
        x,y = None, None
        try:
            x = int(ev.pos().x())
            y = int(ev.pos().y())
            try:
                maskVal  = self.parent.currMask[y,x]
                if maskVal == 0:
                    val = self.parent.currIm[y,x]
                else:
                    val = maskVal
            except IndexError:
                val = None
        except AttributeError:
            val = None
        self.parent.updateDataDisplay(x = x, y = y, val = val)

        if self.parent.in_addregion_stroke:
            self.drawAt(ev.pos())
            if self.is_at_start(ev.pos()):
                self.end_stroke()
                self.parent.in_addregion_stroke = False
        return
    
    def create_start(self, pos):
        self.scatter = pg.ScatterPlotItem([pos.x()], [pos.y()], pxMode=False,
                                        pen=pg.mkPen(color=(255,0,0), width=self.parent.brush_size),
                                        size=max(3*2, self.parent.brush_size*1.8*2), brush=None)
        self.parent.view.addItem(self.scatter)
    
    def is_at_start(self, pos):
        if len(self.parent.currStroke) > 16:
            x0,y0 = self.parent.currStroke[0]
            x1,y1 = pos.x(), pos.y()
            d = (x1-x0)**2 + (y1-y0)**2
            d = math.sqrt(d)
            return d<1
        else:
            return False

    def addSet(self):
        print("adding set")
        if len(self.parent.currStroke)>3:
            print("\tlength requirement")
            ny, nx = self.parent.label().x(), self.parent.label().y()
            img = Image.new('L', (ny, nx), 0)
            ID.Draw(img).polygon(self.parent.currStroke, outline=1, fill=1)
            polygon = np.array(img).astype(bool)
            self.parent.currMask[polygon] = self.colorNum
            self.setImage(self.parent.maskColors[self.parent.currMask], autolevels  = False)
            #self.parent.label().save()
            self.parent.strokes.append((polygon, self.colorNum))
            self.parent.addRecentDrawing()

            if self.parent.contourOn:
                self.parent.drawContours()
    
    def end_stroke(self):
        print("ending stroke")
        self.parent.view.removeItem(self.scatter)
        if not self.strokeAppended:
            # self.parent.addRegionStrokes.append(self.currStroke)
            # self.strokeAppended = True
            # self.currStroke = np.array(self.currStroke)
            # ioutline = self.currStroke[:,3]==1
            # self.parent.currPointSet.extend(list(self.currStroke[ioutline]))
            self.addSet()
            self.parent.currStroke = []
            self.colorNum = 0
            self.parent.in_addregion_stroke = False

    
    def enterEvent(self, ev):
        self.hoverEvent(ev)
    
    def leaveEvent(self, ev):
        self.parent.updateDataDisplay(x = None, y = None, val =None)
        
    
    def paintBrush(self,pos1, pos2):
        print("paint brush")
        pos1y,pos1x = (int(pos1.y()), int(pos1.x()))
        pos2y,pos2x = (int(pos2.y()), int(pos2.x()))

        selem = disk(self.parent.brush_size-2)

        rr,cc,_ = line_aa(pos2y,pos2x, pos1y, pos1x)

        to_change = np.zeros(self.parent.currMask.shape, dtype = bool)
        to_change[rr,cc] = True
        to_change = dilation(to_change, selem)
        self.parent.currMask[to_change] = self.colorNum

        self.parent.strokes.append((to_change, self.colorNum))
        
        self.setImage(self.parent.maskColors[self.parent.currMask])
        #self.parent.label().save()

        if self.parent.drawType == "Outline":
            # for ky,y in enumerate(np.arange(ty[0], ty[1], 1, int)):
            #     for kx,x in enumerate(np.arange(tx[0], tx[1], 1, int)):
            #         iscent = np.logical_and(kx==kcent[0], ky==kcent[1])
            #         self.currStroke.append([self.parent.maskZ, x, y, iscent])
            print("appending stroke")
            self.parent.currStroke.append((int(pos2.x()), int(pos2.y())))
        self.parent.addRecentDrawing()

    
    def drawAt(self, p, ev=None):
        print("drawAt")
        print(self.parent.drawType)
        pos = [int(p.y()), int(p.x())]
        dk = self.drawKernel
        kc = self.drawKernelCenter
        sx = [0,dk.shape[0]]
        sy = [0,dk.shape[1]]
        tx = [pos[0] - kc[0], pos[0] - kc[0]+ dk.shape[0]]
        ty = [pos[1] - kc[1], pos[1] - kc[1]+ dk.shape[1]]
        kcent = kc.copy()
        if tx[0]<=0:
            sx[0] = 0
            sx[1] = kc[0] + 1
            tx    = sx
            kcent[0] = 0
        if ty[0]<=0:
            sy[0] = 0
            sy[1] = kc[1] + 1
            ty    = sy
            kcent[1] = 0
        if tx[1] >= self.parent.label().y()-1:
            sx[0] = dk.shape[0] - kc[0] - 1
            sx[1] = dk.shape[0]
            tx[0] = self.parent.label().y() - kc[0] - 1
            tx[1] = self.parent.label().y()
            kcent[0] = tx[1]-tx[0]-1
        if ty[1] >= self.parent.label().x()-1:
            sy[0] = dk.shape[1] - kc[1] - 1
            sy[1] = dk.shape[1]
            ty[0] = self.parent.label().x() - kc[1] - 1
            ty[1] = self.parent.label().x()
            kcent[1] = ty[1]-ty[0]-1


        ts = (slice(tx[0],tx[1]), slice(ty[0],ty[1]))
        #ss = (slice(sx[0],sx[1]), slice(sy[0],sy[1]))
        self.parent.currMask[ts] = self.colorNum

        self.parent.strokes.append((ts, self.colorNum))

        self.setImage(self.parent.maskColors[self.parent.currMask], autolevels  = False)
        #self.parent.label().save()
        if self.parent.drawType == "Brush":
            print("brush drawtype, adding recent")
            self.parent.addRecentDrawing()
        if self.parent.drawType == "Outline":
            # for ky,y in enumerate(np.arange(ty[0], ty[1], 1, int)):
            #     for kx,x in enumerate(np.arange(tx[0], tx[1], 1, int)):
            #         iscent = np.logical_and(kx==kcent[0], ky==kcent[1])
            #         self.currStroke.append([self.parent.maskZ, x, y, iscent])
            print("appending stroke")
            self.parent.currStroke.append((int(p.x()), int(p.y())))

    
    def addMask(self, mask, color, update = True):
        mask = mask>0
        self.image[mask,:] = color
        if update:
            self.updateImage()

  
class ViewBoxNoRightDrag(pg.ViewBox):
    def __init__(self, parent=None, border=None, lockAspect=False, enableMouse=True, invertY=False, enableMenu=True, name=None, invertX=False):
        pg.ViewBox.__init__(self, None, border, lockAspect, enableMouse,
                            invertY, enableMenu, name, invertX)
        self.parent = parent
        self.axHistoryPointer = -1

    def keyPressEvent(self, ev):
        """
        This routine should capture key presses in the current view box.
        The following events are implemented:
        +/= : moves forward in the zooming stack (if it exists)
        - : moves backward in the zooming stack (if it exists)
        """
        ev.accept()
        if ev.text() == '-':
            self.scaleBy([1.1, 1.1])
        elif ev.text() in ['+', '=']:
            self.scaleBy([0.9, 0.9])
        else:
            ev.ignore()

    def mouseDragEvent(self, ev, axis=None):
        ## if axis is specified, event will only affect that axis.
        if self.parent is None or (self.parent is not None and not self.parent.in_addregion_stroke):
            ev.accept()  ## we accept all buttons

            pos = ev.pos()
            lastPos = ev.lastPos()
            dif = pos - lastPos
            dif = dif * -1

            ## Ignore axes if mouse is disabled
            mouseEnabled = np.array(self.state['mouseEnabled'], dtype=float)
            mask = mouseEnabled.copy()
            if axis is not None:
                mask[1-axis] = 0.0

            ## Scale or translate based on mouse button
            if ev.button() & (QtCore.Qt.LeftButton | QtCore.Qt.MidButton):
                if self.state['mouseMode'] == pg.ViewBox.RectMode:
                    if ev.isFinish():  ## This is the final move in the drag; change the view scale now
                        self.rbScaleBox.hide()
                        ax = QtCore.QRectF(Point(ev.buttonDownPos(ev.button())), Point(pos))
                        ax = self.childGroup.mapRectFromParent(ax)
                        self.showAxRect(ax)
                        self.axHistoryPointer += 1
                        self.axHistory = self.axHistory[:self.axHistoryPointer] + [ax]
                    else:
                        ## update shape of scale box
                        self.updateScaleBox(ev.buttonDownPos(), ev.pos())
                else:
                    tr = dif*mask
                    tr = self.mapToView(tr) - self.mapToView(Point(0,0))
                    x = tr.x() if mask[0] == 1 else None
                    y = tr.y() if mask[1] == 1 else None

                    self._resetTarget()
                    if x is not None or y is not None:
                        self.translateBy(x=x, y=y)
                    self.sigRangeChangedManually.emit(self.state['mouseEnabled'])


import numpy as np
from skimage.morphology import disk, binary_erosion, binary_dilation
import torch
import tensorflow as tf
from skimage.measure import label
import cv2
import skimage
import matplotlib.pyplot as plt
import cv2
from cv2 import resize
import logging
import pathlib
import sys
import os
from PIL import Image

def write_images_to_dir(path, ims, extension = ".tif"):
    dir = os.path.dirname(path)
    if os.path.exists(path):
        return
    os.mkdir(path)
    if "." not in extension:
        extension = "." + extension
    for i, im in enumerate(ims):
        name = f"im_{i}{extension}"
        fname = os.path.join(path, name)
        skimage.io.imsave(fname, im)


def overlay_images(phase_img, mask_img, colorized_mask):
    # Convert numpy arrays to PIL images
    mask = mask_img.copy()
    mask[mask>0] = 255
    microscopy_pil = Image.fromarray(convertGreyToRGB(phase_img))
    mask_pil = Image.fromarray(mask)
    colorized_pil = Image.fromarray(colorized_mask)

    # Convert images to 'RGBA' mode
    microscopy_pil = microscopy_pil.convert("RGBA")
    colorized_pil = colorized_pil.convert("RGBA")
    mask_pil = mask_pil.convert("L")

    # Overlay images
    microscopy_pil.paste(colorized_pil, (0, 0), mask_pil)

    # Convert PIL image back to numpy array
    result = np.array(microscopy_pil)

    return result


def get_frames_with_no_cells(masks):
    no_cell_frames = []
    for i, mask in enumerate(masks):
        if np.all(mask==0):
            no_cell_frames.append(i)
    return no_cell_frames

def get_end_of_cells(masks):
    for i in reversed(range(masks.shape[0])):
        if np.any(masks[i]!=0):
            return i



def get_filename(path):
    _, name_with_extension = os.path.split(path)
    return os.path.splitext(path)[0]

def logger_setup():
    cp_dir = pathlib.Path.home().joinpath('.yeastvision')
    cp_dir.mkdir(exist_ok=True)
    log_file = cp_dir.joinpath('run.log')
    try:
        log_file.unlink()
    except:
        print('creating new log file')
    logging.basicConfig(
                    level=logging.INFO,
                    format="%(asctime)s [%(levelname)s] %(message)s",
                    handlers=[
                        logging.FileHandler(log_file),
                        logging.StreamHandler(sys.stdout)
                    ]
                )
    logger = logging.getLogger(__name__)
    logger.info(f'WRITING LOG OUTPUT TO {log_file}')
    return logger, log_file

def configure_tf_memory_growth():
    physical_devices = tf.config.list_physical_devices('GPU')
    try:
        tf.config.experimental.set_memory_growth(physical_devices[0], True)
    except:
        # Invalid device or cannot modify virtual devices once initialized.
        pass


def check_gpu(do_tf = True, do_torch = True):
    if do_tf:
        print(f"Tensorflow\n__________:\n{tf.config.experimental.get_memory_info('GPU:0')}")
    if do_torch:
        t = torch.cuda.get_device_properties(0).total_memory
        r = torch.cuda.memory_reserved(0)
        a = torch.cuda.memory_allocated(0)
        d = {"total": t,
             "reserved": r,
             "allocated": a}
        print(f"Pytorch\n_________\n{d}")
        # free inside reserved

def showCellNums(mask):
    "annotates the current plt figure with cell numbers"
    #cell_vals = np.unique(mask[mask!=0]).astype(int)
    cell_vals = np.unique(mask).astype(int)
    cell_vals = np.delete(cell_vals, np.where(cell_vals == 0))

    for val in cell_vals:
        # print("cell val: " + str(val)) #test
        x, y = getCenter(mask, val)
        plt.annotate(str(val), xy=(x, y), ha='center', va='center')

def getCenter(mask, cell_val):
    '''
    takes random points within a cell to estimate its center. Used to find the x,y coordinates where the cell number
    annotation text should be displated
    '''
    y, x = (mask == cell_val).nonzero()
    sample = np.random.choice(len(x), size=20, replace=True)
    meanx = np.mean(x[sample])
    meany = np.mean(y[sample])

    return int(round(meanx)), int(round(meany))

def rescaleByFactor(factor, ims):
    row, col = ims[0].shape
    newrow, newcol = int(row*factor), int(col*factor)
    print(newrow, newcol)
    return rescaleBySize((newrow, newcol), ims)

def rescaleBySize(newshape, ims):
    row, col = newshape
    return [resize(im, (col,row), interpolation=cv2.INTER_CUBIC) for im in ims]

def binarize_relabel(mask):
    return label(binarize(mask))

def binarize(mask):
    return (mask > 0).astype(np.uint8)

def check_torch_gpu(gpu_number = 0):
    try:
        device = torch.device('cuda:' + str(gpu_number))
        _ = torch.zeros([1, 2, 3]).to(device)
        print('** TORCH CUDA version installed and working. **')
        return True
    except:
        print('TORCH CUDA version not installed/working.')
        return False

def check_tf_gpu():
    gpus = tf.config.list_physical_devices('GPU')
    
    if gpus:
        print("** TENSORFLOW installed and working. **")
    else:
        print("TENSORFLOW not installed/working")
    return bool(gpus)

def count_objects(labeledMask):
    return len(np.unique(labeledMask))-1

def capitalize(string):
    first_letter = string[0]
    rest_of_string = string[1:]
    return first_letter.capitalize()+rest_of_string

def get_mask_contour(mask):
    out_mask = np.zeros_like(mask)
    for cellnum in np.unique(mask):
        if cellnum>0:
            cell_contour = get_cell_contour(mask, cellnum)
            out_mask[cell_contour==1]=cellnum
    return out_mask

def get_cell_contour(mask, cellnum):
    out_mask = np.zeros_like(mask)
    mask_binary = (mask==cellnum).astype(np.uint8)
    in_mask = shrink_bud(mask_binary, kernel_size = 2)
    mask_binary[in_mask]=0
    return mask_binary


def shrink_bud(bud_mask, footprint = None, kernel_size = 2):
    if not footprint:
        footprint = disk(kernel_size)
    return binary_erosion(bud_mask, footprint)


def enlarge_bud(bud_mask, footprint = None, kernel_size = 4):
    if not footprint:
        footprint = disk(kernel_size)
    return binary_dilation(bud_mask, footprint)

def normalize_im(im_o, clip = True):
    """
    Normalizes a given image such that the values range between 0 and 1.     
    
    Parameters
    ---------- 
    im : 2d-array
        Image to be normalized.
    clip: boolean
        Whether or not to set im values <0 to 0
        
    Returns
    -------
    im_norm: 2d-array
        Normalized image ranging from 0.0 to 1.0. Note that this is now
        a floating point image and not an unsigned integer array. 
    """ 
    im = np.nan_to_num(im_o)
    if clip:
        im[im<0] = 0
    if im.max()==0:
        return im.astype(np.float32)
    im_norm = (im - im.min()) / (im.max() - im.min())
    im_norm[np.isnan(im_o)] = np.nan
    return im_norm.astype(np.float32)

def convertGreyToRGB(im):
    image = skimage.util.img_as_ubyte(normalize_im(im))
    image_3D = cv2.merge((image,image,image))
    #return cv2.cvtColor(image, cv2.COLOR_GRAY2RGBA)
    return image_3D
    # rgba = cv2.cvtColor(rgb, cv2.COLOR_RGB2RGBA)
    # rgba[:, :, 3] = 1
    # return rgba

def get_img_from_fig(fig):
    '''
    input: matplotlib pyplot figure object
    output: a 3D numpy array object corresponding to the image shown by that figure
    '''
    import io as i
    buf = i.BytesIO()
    fig.savefig(buf, format="tif", bbox_inches='tight', pad_inches=0)
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    buf.close()
    img = cv2.imdecode(img_arr, 1)
    img = (cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

    return (np.array(img)).astype("uint8")


def overlay(im, mask, color = [255,0,0], get_fig = False, image_only = True):
    '''
    Overlays an image and with its binary mask for easily visualizing
    
    
    Parameters
    --------
    image: grayscale ndarray of shape mxn
    mask: binary ndaraay of shape mxn
    get_fig (bool): return the matplotlib figure as a np array
    color (list): color given to mask in [R,G,B] form
    image_only (bool): return only the merged images as a RGB ndarray 
    '''
    image = skimage.util.img_as_ubyte(normalize_im(im))
    image_3D = cv2.merge((image,image,image))
    mask_plot = image_3D.copy()
    mask_plot[mask==1] = color # set thresholded pixels to red
    
    if image_only:
        return mask_plot
    
    fig = plt.figure()
    plt.axis('off')
    plt.imshow(image_3D)
    plt.imshow(mask_plot)
    plt.show()
    
    if get_fig:
        return get_img_from_fig(fig)
    else:
        return mask_plot


from PyQt5 import QtGui, QtCore
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import (QApplication,QSlider, QStyle, QStyleOptionSlider, QPushButton,QCheckBox, QComboBox, QFrame)
from PyQt5.QtCore import Qt


class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)


class ReadOnlyCheckBox(QCheckBox):
    def __init__(self, parent=None):
        super().__init__(parent)
    
    def mousePressEvent(self, event):
        event.ignore()


class MaskTypeButtons():
    def __init__(self, parent=None, row=0, col=0):
        super(MaskTypeButtons, self).__init__()
        self.parent = parent
        self.buttonStrings = ["Mask", "Probability"] # Mask id = 0, Probability = 2
        self.buttons = []

        self.maskId = 0
        self.probId = 1

        for colAdd in range(len(self.buttonStrings)):
            button = QCheckBox(self.buttonStrings[colAdd])
            button.setStyleSheet(self.parent.checkstyle)
            button.setFont(self.parent.medfont)

            if colAdd == self.maskId:
                button.stateChanged.connect(self.maskButtonPress)
            else:
                button.stateChanged.connect(self.probButtonPress)

            self.parent.l.addWidget(button,row, col+2*colAdd, 1,2)
            button.setCheckState(False)
            button.setEnabled(False)
            self.buttons.append(button)
        
        self.maskButton = self.buttons[self.maskId]
        self.probButton = self.buttons[self.probId]
    
    def maskButtonPress(self):
        if self.maskButton.isChecked():

            if self.parent.maskOn:
                self.toggleButton(self.probButton, False)
                self.parent.changeMaskType(self.maskId)
            else:
                self.parent.maskType = self.maskId
                self.parent.addMask()
        
        elif not self.maskButton.isChecked():
            self.parent.hideMask()
    
    def checkMaskRemote(self):
        if not self.maskButton.isChecked():
            self.maskButton.setChecked(True)
            self.parent.maskType = self.maskId
    
    def checkProbRemote(self):
        if not self.probButton.isChecked():
            self.probButton.setChecked(True)
    
    def probButtonPress(self):

        if self.parent.maskData.channels[self.parent.maskZ].shape[0]>1:

            if not self.probButton.isChecked():
                self.parent.hideMask()
            
            elif self.probButton.isChecked():
                if self.parent.maskOn:  
                    self.toggleButton(self.maskButton, False)
                    self.parent.changeMaskType(self.probId)
                else:
                    self.parent.maskType = self.probId
        
        else:
            self.probButton.setCheckState(False)

    def setEnabled(self, b):
        for button in self.buttons:
            button.setEnabled(b)
    
    def uncheckAll(self):
        for button in self.buttons:
            button.setCheckState(False)
    
    def toggleButton(self, button, boolean):
        button.blockSignals(True)
        button.setCheckState(boolean)
        button.blockSignals(False)      

class RangeSlider(QSlider):
    """ A slider for ranges.
        This class provides a dual-slider for ranges, where there is a defined
        maximum and minimum, as is a normal slider, but instead of having a
        single slider value, there are 2 slider values.
        This class emits the same signals as the QSlider base class, with the
        exception of valueChanged
        Found this slider here: https://www.mail-archive.com/pyqt@riverbankcomputing.com/msg22889.html
        and modified it
    """
    def __init__(self, parent=None, *args):
        super(RangeSlider, self).__init__(*args)

        self._low = self.minimum()
        self._high = self.maximum()

        self.setTickInterval(1)

        self.pressed_control = QStyle.SC_None
        self.hover_control = QStyle.SC_None
        self.click_offset = 0

        self.setOrientation(QtCore.Qt.Horizontal)
        self.setTickPosition(QSlider.TicksRight)
        self.setStyleSheet(\
                "QSlider::handle:horizontal {\
                background-color: white;\
                border: 1px solid white;\
                border-radius: 2px;\
                border-color: white;\
                height: 8px;\
                width: 6px;\
                margin: 0px 2; \
                }")

        # 0 for the low, 1 for the high, -1 for both
        self.active_slider = 0
        self.parent = parent

    def level_change(self):
        if self.parent is not None:
            self.parent.saturation = [self._low, self._high]
            self.parent.pg_im.setLevels(self.parent.saturation)

    def low(self):
        return self._low

    def setLow(self, low):
        self._low = low
        self.update()

    def high(self):
        return self._high

    def setHigh(self, high):
        self._high = high
        self.update()
    
    def resetLevels(self, saturationList):
        self.setLow(saturationList[0])
        self.setHigh(saturationList[1])

    def paintEvent(self, event):
        # based on http://qt.gitorious.org/qt/qt/blobs/master/src/gui/widgets/qslider.cpp
        painter = QPainter(self)
        style = QApplication.style()

        for i, value in enumerate([self._low, self._high]):
            opt = QStyleOptionSlider()
            self.initStyleOption(opt)

            # Only draw the groove for the first slider so it doesn't get drawn
            # on top of the existing ones every time
            if i == 0:
                opt.subControls = QStyle.SC_SliderHandle#QStyle.SC_SliderGroove | QStyle.SC_SliderHandle
            else:
                opt.subControls = QStyle.SC_SliderHandle

            if self.tickPosition() != self.NoTicks:
                opt.subControls |= QStyle.SC_SliderTickmarks

            if self.pressed_control:
                opt.activeSubControls = self.pressed_control
                opt.state |= QStyle.State_Sunken
            else:
                opt.activeSubControls = self.hover_control

            opt.sliderPosition = int(value)
            opt.sliderValue = int(value)
            style.drawComplexControl(QStyle.CC_Slider, opt, painter, self)

    def mousePressEvent(self, event):
        event.accept()

        style = QApplication.style()
        button = event.button()
        # In a normal slider control, when the user clicks on a point in the
        # slider's total range, but not on the slider part of the control the
        # control would jump the slider value to where the user clicked.
        # For this control, clicks which are not direct hits will slide both
        # slider parts
        if button:
            opt = QStyleOptionSlider()
            self.initStyleOption(opt)

            self.active_slider = -1

            for i, value in enumerate([self._low, self._high]):
                opt.sliderPosition = value
                hit = style.hitTestComplexControl(style.CC_Slider, opt, event.pos(), self)
                if hit == style.SC_SliderHandle:
                    self.active_slider = i
                    self.pressed_control = hit

                    self.triggerAction(self.SliderMove)
                    self.setRepeatAction(self.SliderNoAction)
                    self.setSliderDown(True)

                    break

            if self.active_slider < 0:
                self.pressed_control = QStyle.SC_SliderHandle
                self.click_offset = self.__pixelPosToRangeValue(self.__pick(event.pos()))
                self.triggerAction(self.SliderMove)
                self.setRepeatAction(self.SliderNoAction)
        else:
            event.ignore()

    def mouseMoveEvent(self, event):
        if self.pressed_control != QStyle.SC_SliderHandle:
            event.ignore()
            return

        event.accept()
        new_pos = self.__pixelPosToRangeValue(self.__pick(event.pos()))
        opt = QStyleOptionSlider()
        self.initStyleOption(opt)

        if self.active_slider < 0:
            offset = new_pos - self.click_offset
            self._high += offset
            self._low += offset
            if self._low < self.minimum():
                diff = self.minimum() - self._low
                self._low += diff
                self._high += diff
            if self._high > self.maximum():
                diff = self.maximum() - self._high
                self._low += diff
                self._high += diff
        elif self.active_slider == 0:
            if new_pos >= self._high:
                new_pos = self._high - 1
            self._low = new_pos
        else:
            if new_pos <= self._low:
                new_pos = self._low + 1
            self._high = new_pos

        self.click_offset = new_pos
        self.update()

    def mouseReleaseEvent(self, event):
        self.level_change()

    def __pick(self, pt):
        if self.orientation() == QtCore.Qt.Horizontal:
            return pt.x()
        else:
            return pt.y()


    def __pixelPosToRangeValue(self, pos):
        opt = QStyleOptionSlider()
        self.initStyleOption(opt)
        style = QApplication.style()

        gr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderGroove, self)
        sr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderHandle, self)

        if self.orientation() == QtCore.Qt.Horizontal:
            slider_length = sr.width()
            slider_min = gr.x()
            slider_max = gr.right() - slider_length + 1
        else:
            slider_length = sr.height()
            slider_min = gr.y()
            slider_max = gr.bottom() - slider_length + 1

        return style.sliderValueFromPosition(self.minimum(), self.maximum(),
                                             pos-slider_min, slider_max-slider_min,
                                             opt.upsideDown)

class ModelButton(QPushButton):
    def __init__(self, parent, model_name, text):
        super().__init__()
        self.setEnabled(True)
        self.setCheckable(True)
        self.setStyleSheet(parent.styleCheckable)
        self.setText(model_name)
        self.setToolTip(text)
        self.setFont(parent.smallfont)
        #self.clicked.connect(lambda: self.press(parent))
        self.model_name = model_name
        
    def press(self, parent):
        for i in range(len(parent.StyleButtons)):
            parent.StyleButtons[i].setStyleSheet(parent.styleUnpressed)
        self.setStyleSheet(parent.stylePressed)
        parent.compute_model(self.model_name)


class CheckableComboBox(QComboBox):
    def __init__(self):
        super().__init__()
        self._changed = False

        self.view().pressed.connect(self.handleItemPressed)

    def setItemChecked(self, index, checked=False):
        item = self.model().item(index, self.modelColumn()) # QStandardItem object

        if checked:
            item.setCheckState(Qt.Checked)
        else:
            item.setCheckState(Qt.Unchecked)

    def handleItemPressed(self, index):
        item = self.model().itemFromIndex(index)

        if item.checkState() == Qt.Checked:
            item.setCheckState(Qt.Unchecked)
        else:
            item.setCheckState(Qt.Checked)
        self._changed = True


    def hidePopup(self):
        if not self._changed:
            super().hidePopup()
        self._changed = False

    def itemChecked(self, index):
        item = self.model().item(index, self.modelColumn())
        return item.checkState() == Qt.Checked

    def getCheckedItems(self):
        checkedItems = [self.itemText(i) for i in range(self.count()) if self.itemChecked(i)]
        return checkedItems






    






from PyQt5 import QtGui, QtCore

import pyqtgraph as pg
from PyQt5.QtWidgets import (QApplication, QTreeWidgetItem, QPushButton, QDialog,
                        QDialogButtonBox, QLineEdit, QFormLayout, QCheckBox,  QSpinBox, QDoubleSpinBox, QLabel, 
                            QWidget, QComboBox, QGridLayout, QHBoxLayout, QSizePolicy, QHeaderView, QVBoxLayout,
                            QScrollArea, QTreeWidget)
from PyQt5.QtCore import Qt
import numpy as np
import os
import datetime
from yeastvision.parts.guiparts import *
from yeastvision.plot.plot import PlotProperty

class FigureDialog(QDialog):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.layout = QGridLayout()
        self.row = 0

        self.label_widgets = {}
        self.label_names = self.parent.experiment().get_label_names()
        self.make_top_row()
        for i, label_name in enumerate(self.label_names):
            self.add_row(label_name, i)

    
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.layout.addWidget(self.buttonBox, self.row, 0)
        
        self.setLayout(self.layout)

    def make_top_row(self):
        for i, header in enumerate(["label name", "contour", "full label", "order"]):
            self.layout.addWidget(QLabel(header), self.row, i,1,1)
        self.row+=1
    
    def add_row(self, label_name, i):
        self.label_widgets[label_name] = {}
        self.layout.addWidget(QLabel(label_name), self.row, 0,1,1)

        contour_check = QCheckBox()
        self.label_widgets[label_name]["contours"] = contour_check
        self.layout.addWidget(contour_check, self.row,1,1,1)

        label_check = QCheckBox()
        self.label_widgets[label_name]["labels"] = label_check
        self.layout.addWidget(label_check, self.row,2,1,1)

        order_spin = QSpinBox()
        order_spin.setMaximum(len(self.label_names)-1)
        order_spin.setValue(i)
        self.label_widgets[label_name]["order"] = order_spin
        self.layout.addWidget(order_spin, self.row, 3,1,1)

        self.row+=1
    
    def get_data(self):
        data = self.label_widgets.copy()
        for name in data.keys():
            data[name]["contours"] = data[name]["contours"].isChecked()
            data[name]["labels"] = data[name]["labels"].isChecked()
            data[name]["order"] = data[name]["order"].value()
        return data


class DirReaderDialog(QDialog):
    def __init__(self, dir, fnames):
        super().__init__()

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.formLayout = QFormLayout()
        self.formLayout.addRow(QLabel(f"Load Custom Channels\n{dir}"))
        self.setLayout(self.formLayout)


        # Create the QSpinBox
        self.spinBox = QSpinBox(self)
        self.spinBox.setMinimumWidth(100)
        self.spinBox.setMinimum(0) # Set minimum value
        self.spinBox.valueChanged.connect(self.update_line_edits) # Connect signal to slot
        self.formLayout.addRow("number of channels:", self.spinBox)
        self.lineEditLayout = QFormLayout()
        self.formLayout.addRow(self.lineEditLayout)
        self.formLayout.addWidget(self.buttonBox)

        # Initialize list to hold QLineEdits
        self.line_edits = []
    

    def update_line_edits(self, value):
        # Check if the value has increased or decreased
        if value > len(self.line_edits):
            # Value increased, create new QLineEdit(s)
            for i in range(len(self.line_edits), value):
                line_edit = QLineEdit(self)
                line_edit.setToolTip("a sequence of characters only found in this channels filename")
                self.line_edits.append(line_edit)
                self.lineEditLayout.addRow(f"channel {i} identifier", line_edit)
        elif value < len(self.line_edits):
            # Value decreased, remove QLineEdit(s)
            for i in range(len(self.line_edits)-1, value-1, -1):
                self.lineEditLayout.removeRow(self.line_edits[i]) # Remove QLineEdit from layout
                self.line_edits[i].deleteLater() # Delete QLineEdit from memory
                self.line_edits.pop(i) # Remove QLineEdit from list
    
    def getData(self):
        return [line_edit.text() for line_edit in self.line_edits]
    
class GeneralParamDialog(QDialog):
    def __init__(self, hyperparamDict, paramtypes, windowName, parent, labelSelects = None, channelSelects = None):
        
        super().__init__(parent)
        self.parent = parent
        self.params = hyperparamDict
        self.types = paramtypes

        self.dropDowns = {}
        self.checkBoxes = {}
        self.lineEdits = {}
        self.spinBoxes = {}
        self.dSpinBoxes = {}

        if self.params:
            self.splitHyperParams()

        self.labelSelects = labelSelects
        self.channelSelects = channelSelects

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.message = QLabel(f"Enter parameters for {windowName}")

        self.formLayout = QFormLayout()
        self.formLayout.addRow(self.message)
        self.populateFormLayout()
        self.formLayout.addWidget(self.buttonBox)
        self.setLayout(self.formLayout)
        self.errorAlrShown = False
    
    def splitHyperParams(self):
        for paramName, paramType in zip(list(self.params.keys()),self.types) :
            if paramType == "drop":
                self.dropDowns[paramName] = ""
            elif paramType is bool:
                self.checkBoxes[paramName] = self.params[paramName]
            elif paramType is None:
                self.lineEdits[paramName] = self.params[paramName]
            elif paramType is int:
                self.spinBoxes[paramName] = self.params[paramName]
            elif paramType is float:
                self.dSpinBoxes[paramName] = self.params[paramName]
    
    def populateFormLayout(self):
        self.dropDownData  = {}
        if self.channelSelects:
            for channelSelectName in self.channelSelects:
                self.dropDownData[channelSelectName] = self.addParentComboBox(self.parent.channelSelect, name = channelSelectName)
        if self.labelSelects:
            for labelSelectName in self.labelSelects:
                self.dropDownData[labelSelectName] = self.addParentComboBox(self.parent.labelSelect, name = labelSelectName)
        
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)

        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
        
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setValue(float(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)

    
    def addParentComboBox(self, parentComboBox, name = "Channel", addNA = False):
        channelSelect = QComboBox()
        channelSelect.setFixedWidth(120)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        self.formLayout.addRow(self.produceLabel(name), channelSelect)
        return channelSelect

    def produceLabel(self, text):
        label = QLabel(text)
        return label
    
    def getData(self):
        dropDownData = {k:v.currentText() for k,v in self.dropDownData.items()}
        lineEditData = {k:float(v.text()) for k,v in self.lineEditData.items()}
        boolData = {k:v.isChecked() for k,v in self.checkBoxData.items()}
        spinData = {k:v.value() for k,v in self.spinBoxData.items()}
        return dropDownData | lineEditData | boolData | spinData



class ModelParamDialog(GeneralParamDialog):
    def __init__(self, hyperparamDict, paramtypes, modelName, parent = None):
        super().__init__(hyperparamDict, paramtypes, modelName, parent)

    def populateFormLayout(self):
        self.dropDownData  = {}
        self.dropDownData["Channel"] = self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        # order is important here as self.dropDownData["T Start"]/["T Stop"]
            # must exist before addTValues and indexChange can be called
        self.addTDropDown()
        self.dropDownData["Channel"].currentIndexChanged.connect(self.channelSelectIndexChanged)
        self.channelSelectIndexChanged()
        
        if self.dropDowns:
            for dropDownName in self.dropDowns.keys():
                self.dropDownData[dropDownName] = self.addParentComboBox(self.parent.channelSelect, name = dropDownName)
        
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)

        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
    
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setValue(float(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)

    def channelSelectIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["Channel"].currentText())
        maxT = self.parent.experiment().channels[channelIndex].max_t()
        self.dropDownData["T Start"].clear()
        self.dropDownData["T Stop"].clear()
        self.addTValues(maxT)

    def addTimeDropDown(self, name):
        timeSelect = QComboBox()
        timeSelect.setFixedWidth(120)
        self.formLayout.addRow(self.produceLabel(name), timeSelect)
    
    def addTDropDown(self):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)
        self.dropDownData["T Start"] = timeSelectStart
        self.dropDownData["T Stop"] = timeSelectStop
        self.formLayout.addRow(QLabel("Time Start/Stop"), hbox)
    
    def addTValues(self, maxT, prefix = "T "):
        validTs = list(range(maxT+1))
        validTs = [str(t) for t in validTs]
        self.dropDownData[f"{prefix}Start"].addItems(validTs)
        self.dropDownData[f"{prefix}Start"].setCurrentIndex(self.parent.tIndex)
        self.dropDownData[f"{prefix}Stop"].addItems(validTs)
        self.dropDownData[f"{prefix}Stop"].setCurrentIndex(self.parent.tIndex)
    
    def showErrorMessage(self):
        if not self.errorAlrShown:
            self.message.setText(self.message.text() + "\nEnsure T Start<T Stop and All Line Entries are Numeric")
            self.errorAlrShown = True

    def accept(self):
        if self.validateLineData() and self.validateTData():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTData(self):
        data = self.dropDownData
        valid = int(data["T Start"].currentText())<=int(data["T Stop"].currentText())
        return valid

    def validateLineData(self):
        data = {k:v.text() for k,v in self.lineEditData.items()}
        i = 0
        vals = list(data.values())
        valid = True
        while i<len(vals) and valid:
            valid = self.isNumber(vals[i]) and float(vals[i])>=0.0
            i+=1
        return valid
    
    def isNumber(self, num):
        """ Returns True is string is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()


class FlouParamDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.message = QLabel(f"Parameters for blob detection")

        self.formLayout = QFormLayout()
        self.formLayout.addRow(self.message)
        self.populateFormLayout()
        self.formLayout.addWidget(self.buttonBox)
        self.setLayout(self.formLayout)
        self.errorAlrShown = False
    
    def populateFormLayout(self):
        self.checkBoxData = {}
        useExampleCheck = QPushButton("extract params from current mask")
        useExampleCheck.setEnabled(np.any(self.parent.getCurrMaskSet() > 0))
        useExampleCheck.setCheckable(True)
        useExampleCheck.setChecked(False)
        useExampleCheck.clicked.connect(self.exampleCheckClicked)
        self.checkBoxData["extract from curr mask"] = useExampleCheck
        self.formLayout.addWidget(useExampleCheck)

        self.dropDownData = {}
        self.dropDownData["channel"] = self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        self.dropDownData["label"] = self.addParentComboBox(self.parent.labelSelect, name = "Filter Masks")
        self.dropDownData["label"].addItem("")
        self.dropDownData["label"].setCurrentIndex(-1)
        self.dropDownData["label"].setToolTip("Anything segmented outside 'Filter Mask' will be set to 0")
        contourCheck = QPushButton("use contours")
        contourCheck.setEnabled(np.any(self.parent.getCurrMaskSet() > 0))
        contourCheck.setCheckable(True)
        if contourCheck.isCheckable:
            contourCheck.setChecked(self.parent.contourOn)
        self.formLayout.addWidget(contourCheck)
        self.checkBoxData["use contours"] = contourCheck

        self.addTDropDown()
        self.dropDownData["channel"].currentIndexChanged.connect(self.dIndexChanged)
        self.dropDownData["label"].currentIndexChanged.connect(self.dIndexChanged)
        self.dIndexChanged()

        self.lineEditData = {}
        for lineEditName, defaultVal in zip(["min area", "max area", "threshold"], [20,50,0.08]):
            lineEdit = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[lineEditName] = lineEdit
            self.formLayout.addRow(self.produceLabel(lineEditName), lineEdit)

    def exampleCheckClicked(self):
        enabled = not self.checkBoxData["extract from curr mask"].isChecked()
        for checkBox in self.lineEditData.values():
            checkBox.setEnabled(enabled)

    def dIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["channel"].currentText())
        labelIndex = self.parent.labelSelect.findText(self.dropDownData["label"].currentText())

        if labelIndex == -1:
            maxT = self.parent.imData.maxTs[channelIndex]
        else:
            maxT = min(self.parent.imData.maxTs[channelIndex], self.parent.maskData.maxTs[labelIndex])
        
        self.dropDownData["T Start"].clear()
        self.dropDownData["T Stop"].clear()
        self.addTValues(maxT)

    def addParentComboBox(self, parentComboBox, name = "Channel", addNA = False):
        channelSelect = QComboBox()
        channelSelect.setFixedWidth(120)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        self.formLayout.addRow(self.produceLabel(name), channelSelect)
        return channelSelect

    def produceLabel(self, text):
        label = QLabel(text)
        return label
    
    def getData(self):
        dropDownData = {k:v.currentText() for k,v in self.dropDownData.items()}
        lineEditData = {k:float(v.text()) for k,v in self.lineEditData.items()}
        boolData = {k:v.isChecked() for k,v in self.checkBoxData.items()}
        return dropDownData | lineEditData | boolData

    def addTimeDropDown(self, name):
        timeSelect = QComboBox()
        timeSelect.setFixedWidth(120)
        self.formLayout.addRow(self.produceLabel(name), timeSelect)
    
    def addTDropDown(self):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)
        self.dropDownData["T Start"] = timeSelectStart
        self.dropDownData["T Stop"] = timeSelectStop
        self.formLayout.addRow(QLabel("Time Start/Stop"), hbox)
    
    def addTValues(self, maxT, prefix = "T "):
        validTs = list(range(maxT+1))
        validTs = [str(t) for t in validTs]
        self.dropDownData[f"{prefix}Start"].addItems(validTs)
        self.dropDownData[f"{prefix}Start"].setCurrentIndex(0)
        self.dropDownData[f"{prefix}Stop"].addItems(validTs)
        self.dropDownData[f"{prefix}Stop"].setCurrentIndex(len(validTs)-1)

    def showErrorMessage(self):
        if not self.errorAlrShown:
            self.message.setText(self.message.text() + "\nEnsure T Start<T Stop and All Line Entries are Numeric")
            self.errorAlrShown = True

    def accept(self):
        if self.validateLineData() and self.validateTData():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTData(self):
        data = self.dropDownData
        valid = int(data["T Start"].currentText())<=int(data["T Stop"].currentText())
        return valid

    def validateLineData(self):
        data = {k:v.text() for k,v in self.lineEditData.items()}
        i = 0
        vals = list(data.values())
        valid = True
        while i<len(vals) and valid:
            valid = self.isNumber(vals[i]) and float(vals[i])>=0.0
            i+=1
        return valid
    
    def isNumber(self, num):
        """ Returns True is string is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()

class ArtilifeParamDialog(ModelParamDialog):
    def __init__(self, params, paramtypes, name, parent):
        super().__init__(params,paramtypes,"artilife", parent)
    
    def populateFormLayout(self):
        self.dropDownData  = {}
        
        self.addArtilifeWeightDropDown()

        self.dropDownData["Channel"] =  self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        # order is important here as self.dropDownData["T Start"]/["T Stop"]
            # must exist before addTValues and indexChange can be called
        self.addTDropDown()
        self.dropDownData["Channel"].currentIndexChanged.connect(self.channelSelectIndexChanged)
        if self.dropDowns:
            for dropDownName in self.dropDowns.keys():
                self.dropDownData[dropDownName] = self.addParentComboBox(self.parent.channelSelect, name = dropDownName, addNA = True)


        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
    
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setValue(float(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        
        self.addArtilifeTDropDowns(name = "mat")
        self.addArtilifeTDropDowns(name = "spore")
        self.channelSelectIndexChanged()
    
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)
        
        self.connectSporeMatButtons()

    def connectSporeMatButtons(self):
        self.checkBoxData["Mating Cells"].toggled.connect(self.toggleMat)
        self.checkBoxData["Sporulating Cells"].toggled.connect(self.toggleSpore)
    
    def toggleMat(self):
        enabled = self.checkBoxData["Mating Cells"].isChecked()
        self.dropDownData["matStart"].setEnabled(enabled)
        self.dropDownData["matStop"].setEnabled(enabled)
        self.dropDownData["matSeg"].setEnabled(enabled)
    def toggleSpore(self):
        enabled = self.checkBoxData["Sporulating Cells"].isChecked()
        self.dropDownData["sporeStart"].setEnabled(enabled)
        self.dropDownData["sporeStop"].setEnabled(enabled)
        self.dropDownData["tetradSeg"].setEnabled(enabled)

    def addArtilifeWeightDropDown(self):
        artilifeWeights = self.parent.getModelWeights(name = "artilife")
        artiWeights = QComboBox()
        artiWeights.addItems(artilifeWeights)
        self.dropDownData["artiWeights"] = artiWeights
        artiWeights.setFixedWidth(120)
        artiWeights.setEnabled(True)
        artiWeights.setCurrentIndex(0)
        self.formLayout.addRow(QLabel("artilife weights"), artiWeights)

    def addArtilifeTDropDowns(self, name):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setEnabled(False)
        timeSelectStop.setEnabled(False)
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)
        self.dropDownData[f"{name}Start"] = timeSelectStart
        self.dropDownData[f"{name}Stop"] = timeSelectStop
        longName = "Mating" if name == "mat" else "Sporulating"
        self.formLayout.addRow(QLabel(f"{longName} Start/Stop"), hbox)
        
        # add dropdown to allow for 
        modelType = "matSeg" if name == "mat" else "tetradSeg"
        modeloptions = [model for model in self.parent.modelNames if modelType in model]
        weights = QComboBox()
        weights.setEnabled(False)
        weights.setFixedWidth(120)
        weights.setToolTip(f"Weights for {modelType}")
        weights.addItems(modeloptions)
        weights.setCurrentIndex(0)
        self.dropDownData[modelType] = weights
        self.formLayout.addRow(QLabel(f"{modelType} Weights"), weights)


    def channelSelectIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["Channel"].currentText())
        maxT = self.parent.experiment().channels[channelIndex].max_t()
        for prefix in ["T ", "mat", "spore"]:
            self.dropDownData[f"{prefix}Start"].clear()
            self.dropDownData[f"{prefix}Stop"].clear()
            self.addTValues(maxT, prefix = prefix)

class TrainWindow(QDialog):
    def __init__(self, modelType, weightPath, parent):
        super().__init__(parent)
        self.setGeometry(100,100,900,350)
        self.setWindowTitle("Train Settings")
        
        self.parent = parent
        self.modelType = modelType
        self.weightPath = weightPath
        self.modelName = self.getModelName(self.weightPath)

        self.win = QWidget(self)
        self.l0 = QGridLayout()
        self.win.setLayout(self.l0)

        self.trainingParamVals, self.lossName = self.getTrainingParams()

        self.yoff = 0
        qlabel = QLabel('Train Model w/ Images')
        qlabel.setFont(QtGui.QFont("Arial", 14, QtGui.QFont.Bold))
        qlabel.setAlignment(QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,2)

        # choose initial model
        self.yoff+=1
        self.modelChoose = QComboBox()
        #self.modelChoose.addItems(modelNames)
        self.modelChoose.addItems([self.modelName], ) 
        self.modelChoose.setFixedWidth(150)
        #self.modelChoose.setCurrentIndex(self.parent.modelChoose.currentIndex())
        self.l0.addWidget(self.modelChoose, self.yoff, 1,1,1)
        qlabel = QLabel('initial model: ')
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)

        # add check box for from scratch training
        self.yoff+=1
        qlabel = QLabel(f"From Scratch")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.checkbox = QCheckBox()
        self.l0.addWidget(qlabel, self.yoff,0,1,1)
        self.l0.addWidget(self.checkbox,self.yoff,1,1,1)

        # add loss function
        self.yoff+=1
        lossDisplay = QComboBox()
        lossDisplay.addItem(str(self.lossName))
        lossDisplay.setFixedWidth(150)
        self.l0.addWidget(lossDisplay, self.yoff,1,1,1)
        qlabel = QLabel("Loss Function: ")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)

        # # select channels for training
        # imSelect = self.addImSelect("Images", self.parent.channelSelect)
        # labelSelect = self.addImSelect("Labels", self.parent.labelSelect)
        # for cSelect in [imSelect, labelSelect]:
        #     self.yoff+=1
        #     cSelect.setFixedWidth(150)
        #     self.l0.addWidget(cSelect.labelObj, self.yoff, 0,1,1)
        #     self.l0.addWidget(cSelect, self.yoff, 1,1,1)

        self.yoff+=1
        qlabel = QLabel("New Model Suffix")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)
        self.suffixName = QLineEdit()
        self.suffixName.setToolTip("Initial model name will automatically be added as a prefix to New Model Suffix.")
        self.suffixName.setText(self.parent.trainModelSuffix)
        self.suffixName.setFixedWidth(200)
        self.l0.addWidget(self.suffixName, self.yoff, 1,1,1)

        self.addTrainParams()
        
        self.addImNames()

        # click button
        self.yoff+=3
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.l0.addWidget(self.buttonBox, self.yoff, 0, 1,4)
        
    
    def accept(self):
        if self.validateTrainingParamVals():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTrainingParamVals(self):
        data = self.edits.values()
        for entry in data:
            if not self.isNumber(entry.text()):
                return False
        return True
    
    def getData(self):
        suffix = str(self.suffixName.text())
        return {k:float(v.text()) for k,v in self.edits.items()} | {"scratch": self.checkbox.isChecked()} | {"model_name": f"{self.modelName}_{suffix}"}

    
    def isNumber(self, num):
        """ Returns True is stsring is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()

    def getModelName(self, weights):
        return os.path.split(weights)[-1].split(".")[0]

    def addImSelect(self, name, parentComboBox):
        parentComboBox = self.parent.channelSelect
        channelSelect = QComboBox()
        channelSelect.setFont(self.parent.medfont)
        channelSelect.setFixedWidth(120)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        channelSelect.labelObj = QLabel(name)
        return channelSelect
    
    def getTrainingParams(self):
        modelClass =  self.parent.getModelClass(self.modelType)
        modelParams = dict(modelClass.trainparams)
        loss = modelClass.loss
        return dict(modelParams), loss
    
    def addTrainParams(self):
        # choose parameters        
        labels = list(self.trainingParamVals.keys())
        defaultVals = list(self.trainingParamVals.values())
        self.edits = {}
        self.yoff += 1
        for i, label in enumerate(labels):
            qlabel = QLabel(label)
            qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            self.l0.addWidget(qlabel, i+self.yoff,0,1,1)
            newEdit = (QLineEdit())
            newEdit.setText(str(defaultVals[i]))
            newEdit.setFixedWidth(200)
            self.edits[label] = newEdit
            self.l0.addWidget(newEdit, i+self.yoff, 1,1,1)
    
    def addImNames(self):
         # list files in folder
        qlabel = QLabel('ims')
        qlabel.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.l0.addWidget(qlabel, 0,4,1,1)
        qlabel = QLabel('ROIs')
        qlabel.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.l0.addWidget(qlabel, 0,5,1,1)
        
        for i in range(10):
            if i > len(self.parent.trainFiles) - 1:
                break
            elif i==9 and len(self.parent.trainFiles) > 10:
                label = '...'
                nmasks = '...'
            else:
                label = os.path.split(self.parent.trainFiles[i])[-1]
                nmasks = str(self.parent.trainLabels[i].max())
            qlabel = QLabel(label)
            self.l0.addWidget(qlabel, i+1,4,1,1)
            qlabel = QLabel(nmasks)
            qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            self.l0.addWidget(qlabel, i+1, 5,1,1)



class PlotWindowCustomize(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setGeometry(100,100,1000,450)
        self.setWindowTitle("Customize Plot Display")
        self.win = QWidget(self)
        self.l = QGridLayout()
        self.win.setLayout(self.l)

        self.morphProps = self.parent.getCellDataLabelProps()
        self.trees = []
        self.boxes = {}
        self.produceParamTreeSingleCell()
        self.produceParamTreeIncludePopulations()

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.l.addWidget(self.buttonBox, 3, 0, 1, 2)

        self.win.show()
    
    def getProps(self, populationObj, propType):
        if propType == "morphology":
            return self.morphProps
        else:
            return [prop for prop in populationObj.properties if prop not in self.morphProps]

    def produceParamTreeSingleCell(self):
        populations = self.parent.getLabelsWithPopulationData()      
        tree = self.produceTree()
        self.trees.append(tree)
        level1 = "single"
        item = QTreeWidgetItem(tree, [level1])
        item.setExpanded(True)

        for level2 in populations:
            setName = self.parent.getTimeSeriesDataName(level2)
            item2 = QTreeWidgetItem(item, [setName])
            item2.setExpanded(True)

            for level3 in ["morphology", "pixel"]:
                item3 = QTreeWidgetItem(item2, [level3])

                for level4 in self.getProps(level2, level3):
                    item4 = QTreeWidgetItem(item3, [level4])
                    box = QCheckBox()
                    tree.setItemWidget(item4, 1, box)
                    self.boxes[PlotProperty(self.parent, "single", setName, level4)] = box
        self.l.addWidget(tree, 0,0,2,1)

    def produceParamTreeIncludePopulations(self):
        populations = self.parent.getLabelsWithPopulationData()      
        for i,level1 in enumerate(self.parent.populationPlotTypes):
            tree = self.produceTree()
            self.trees.append(tree)
            item = QTreeWidgetItem(tree, [level1])
            item.setExpanded(True)

            for level2 in populations:
                setName = self.parent.getTimeSeriesDataName(level2)
                item2 = QTreeWidgetItem(item, [setName])
                item2.setExpanded(True)

                for level2a in level2.population_names:
                    item2a = QTreeWidgetItem(item2, [level2a])

                    for level3 in ["morphology", "pixel"]:
                        item3 = QTreeWidgetItem(item2a, [level3])

                        for level4 in self.getProps(level2, level3):
                            item4 = QTreeWidgetItem(item3, [level4])
                            box = QCheckBox()
                            tree.setItemWidget(item4, 1, box)
                            self.boxes[PlotProperty(self.parent, level1, setName, level4, populationName=level2a)] = box
            self.l.addWidget(tree, 0,i+1,2,1)

    def getData(self):
        checked = [name for name, box in self.boxes.items() if box.isChecked()]
        toPlotDict = {}
        for plot in self.parent.plotTypes:
            toPlotDict[plot] = [prop_obj for prop_obj in checked if plot in prop_obj.plotType]
        return toPlotDict

    
    def produceTree(self):
        tree = QTreeWidget()
        tree.setColumnCount(2)
        tree.setHeaderHidden(True)
        tree.setMinimumHeight(300)
        tree.setMinimumWidth(300)
        tree.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        return tree








from skimage.measure import regionprops, regionprops_table
from skimage.measure import shannon_entropy as shentr
import numpy as np
import pandas as pd
from yeastvision.utils import normalize_im
from tqdm import tqdm

class Cell():
    def __init__(self, mask, num):
        self.cell = (mask == num)
        self.centroid = self.getCentroid()

    def getCentroid(self):
        return regionprops(self.cell.astype(np.uint8))[0].centroid

def getBirthFrame(trackedMasks, cellVal):
    return np.min((np.nonzero(trackedMasks == cellVal))[0])

def getDeathFrame(trackedMasks, cellVal):
    return np.max((np.nonzero(trackedMasks == cellVal))[0])

def entropy(label, intensity_im):
    return shentr(intensity_im, base = 2)

def var(label, intensity_im):
    return np.var(intensity_im)

def median(label, intensity_im):
    return np.median(intensity_im)

def intensity_total(label, intensity_im):
    return np.sum(intensity_im)

def concentration(label, intensity_im):
    return intensity_total(label, intensity_im)/np.sum(label)

LABEL_PROPS = ["area", "eccentricity", "label", ]
IM_PROPS = ["intensity_mean"]
EXTRA_IM_PROPS = [ var, intensity_total, concentration]

def getLifeData(labels):
    lifeData = {"birth":[], "death":[]}
    for val in tqdm(np.unique(labels)[1:]):
        lifeData["birth"].append(getBirthFrame(labels, val))
        lifeData["death"].append(getDeathFrame(labels, val))
    return lifeData

def getPropDf(data):
    #newData = data.drop[data[data["labels"] == "population"].index]
    try:
        return data.drop(columns = ["labels", "birth", "death"])
    except:
        return data.drop(columns = ["labels"])

def exportCellData(data, lifedata):
    print(data)
    export = {"cell":[], "time":[]}
    propDf = getPropDf(data)
    for column in propDf.columns:
            export[column] = []
    
    for cell in data["labels"]:
        firstT = int(lifedata[lifedata["cell"] == cell]["birth"])
        endT = int(lifedata[lifedata["cell"] == cell]["death"])

        print("cell", cell, firstT, endT)

        for i in range(firstT, endT+1):
            export["time"].append(i)
            export["cell"].append(cell)

        for props in propDf.columns:
            vals = data[data["labels"] == cell]
            vals = vals[props].tolist()[0][firstT:endT+1]
            print(cell, props, vals)
            for prop in vals:
                export[props].append(prop)
    print(export)
    return pd.DataFrame(data = export)

def getDaughterMatrix(self, lineageDF):
    arraysize = len(lineageDF.index.tolist())
    daughterMatrix = np.zeros((arraysize, arraysize))
    return daughterMatrix

def getHeatMaps(data):
    heatMaps = []
    for prop in getPropDf(data).columns:
        heatMaps.append(np.array(data[prop].tolist()))
    return np.array(heatMaps)

def getCellData(labels, intensity_ims = None, intensity_im_names = None):
    total_labels = np.unique(labels) # all indeces of cells present in the movie
    total_labels = total_labels[total_labels!=0]
    total_data = {"labels": total_labels} # dictionary storing cell data, to be converted to pd.df
    
    # add birth and death data
    total_data = total_data #| getLifeData(labels)

    # determine whether or not to include image properties in call to regionprops_table
    if intensity_ims:
        intensity_ims = [normalize_im(im) for im in intensity_ims]
        props = LABEL_PROPS + IM_PROPS
        extra_properties = EXTRA_IM_PROPS
    else:
        props = LABEL_PROPS
        extra_properties = None

    # iterate through each time point in given labels (masks)
    for i in tqdm(range(labels.shape[0])):
        label = labels[i,:,:]

        if intensity_ims:
            if len(intensity_ims)>1:
                intensity_im = np.stack([im[i] for im in intensity_ims], -1)
            else:
                intensity_im = intensity_ims[0][i]
        else:
            intensity_im = None
        
        im_data = regionprops_table(label, intensity_im, properties = props, extra_properties = extra_properties)
        
        # add zeros for cells not present in im data
        for label in total_labels:
            if label not in im_data["label"]:
                for property in im_data:
                    vals = [round(val, 3) if val is not None else None for val in list(im_data[property])]
                    im_data[property] = vals

                    if property != "label":
                        # prevent index out of bound error
                        if label-1 < len(im_data[property]):
                            im_data[property][label-1] = None
                        else:
                            im_data[property].append(None)
        
        for property in im_data:
            if property != "label":
                if property not in total_data:
                    total_data[property] = [[prop] for prop in im_data[property]]
                else:
                    for label in total_labels:
                        total_data[property][label-1].append(im_data[property][label-1])
        
    if intensity_im_names:
        for property in im_data:
            isImProp = False
            for prop in IM_PROPS+EXTRA_IM_PROPS:
                try:
                    prop = prop.__name__
                except AttributeError:
                    prop = str(prop)

                if prop in property:
                    isImProp = True
                    break
            if isImProp:
                if "-" in property:
                    idx = int(property.split("-")[-1])-1
                else:
                    idx = 0
                newPropertyName = "-".join([intensity_im_names[idx], prop])
                total_data[newPropertyName] = total_data[property]
                del total_data[property]


                

    
    # # get population average for each property
    # for property in total_data:
    #     if property not in ["labels", "birth", "death"]:
    #         property_array = np.array(total_data[property], dtype = np.float32) # rows are cells (axis 0), columns are time (axis 1)
    #         property_array[property_array == 0.0] = np.nan
    #         total_data[property].append(list(np.nanmean(property_array, axis = 0)))
    #     elif property == "labels":
    #         total_data["labels"] = list(total_data["labels"])
    #         total_data["labels"].append("population")
    #     else:
    #         total_data[property].append(0)

    df = pd.DataFrame.from_dict(total_data)
    return df



from cellpose import models
from yeastvision.models.model import Model as CustomModel
import cv2
import numpy as np
from cellpose.transforms import normalize99
import torch


class CustomCellpose(models.Cellpose):
    def __init__(self, gpu=True, model_type='cyto', pretrained_model = None, net_avg=False, device=None):
        super(models.Cellpose, self).__init__()
        self.torch = True
        
        # assign device (GPU or CPU)
        sdevice, gpu = models.assign_device(self.torch, gpu)
        self.device = device if device is not None else sdevice
        self.gpu = gpu
        print(device, gpu)
        
        model_type = 'cyto' if pretrained_model else None
        
        self.diam_mean = 30. #default for any cyto model 
        nuclear = 'nuclei' in model_type
        if nuclear:
            self.diam_mean = 17. 
        
        self.cp = models.CellposeModel(device=self.device, gpu=self.gpu,
                                pretrained_model= pretrained_model,
                                model_type=None,
                                diam_mean=self.diam_mean,
                                net_avg=net_avg)
        self.cp.model_type = model_type
        
        self.pretrained_size = models.size_model_path(model_type, self.torch)

        self.sz = models.SizeModel(device=self.device, pretrained_size=self.pretrained_size,
                            cp_model=self.cp)
        self.sz.model_type = model_type


class CustomCPWrapper(CustomModel):
    hyperparams  = {  
    "Mean Diameter":0, 
    "Flow Threshold":0.4, 
    "Cell Probability Threshold": 0,
    "Speed mode": False}
    types = [None, None,None,bool]
    
    loss = "categorical crossentropy"
    trainparams = {
                "learning_rate":0.1,
                "weight_decay": 0.0001,
                "n_epochs":100}

    def __init__(self, params, weights):
        super().__init__(params, weights)
        if self.params["Mean Diameter"] == 0:
            self.params["Mean Diameter"] = None
        
        self.model = CustomCellpose(pretrained_model=self.weights)
        self.cpAlone = self.model.cp
        
    
    def processProbability(self, rawProb):
        return (np.clip(normalize99(rawProb.copy()), 0, 1) * 255).astype(np.uint8)

    
    def train(self, ims, labels, params):
        # ims must be 3D for cellpose
        ims = [cv2.merge((im,im,im)) for im in ims]
        # call train the models.CellPoseModel
        self.model.cp.train(ims, labels, 
                                        channels=[0,0], 
                                        save_path=params["dir"], 
                                        min_train_masks=1,
                                        nimg_per_epoch=8,
                                        learning_rate = params['learning_rate'], 
                                        weight_decay = params['weight_decay'], 
                                        n_epochs = int(params['n_epochs']),
                                        model_name = params["model_name"])

    @classmethod
    @torch.no_grad()
    def run(cls, ims, params, weights):
        params = params if params else cls.hyperparams

        model = cls(params, weights)
        ims3D = [cv2.merge((im,im,im)) for im in ims]
        assert len(ims3D[0].shape)==3

        if not params["Mean Diameter"]:
            evaluator = model.cpAlone
            model.masks, flows, _ = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        else:
            evaluator = model.model
            model.masks, flows, _, model.diams = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        model.cellprobs = [flow[2] for flow in flows]
        model.cellprobs = np.array((model.processProbability(model.cellprobs)), dtype = np.uint8)

        masks = np.array(model.masks, dtype = np.uint16)
        prob = model.cellprobs

        del model
        return masks, prob


import os
import numpy as np
from skimage.exposure import equalize_adapthist
from skimage.measure import label
from yeastvision.models.budNET.model import BudNET
from yeastvision.models.utils import prediction
from yeastvision.models.loss import jaccard_coef_loss, jaccard_coef
import matplotlib.pyplot as plt
from tensorflow.keras.models import load_model


class VacNET(BudNET):
    def __init__(self, params, weights):
        super().__init__(params, weights)
    
    


import os
import numpy as np
from skimage.exposure import equalize_adapthist
from .segment import segment
import skimage
from yeastvision.models.model import Model as CustomModel
import matplotlib.pyplot as plt
from .unet import unet

class YeaZ(CustomModel):
    hyperparams = {"Threshold":0.50, "Minimum Cell Distance":5}
    trainparams = {"learning_rate": 0.0001, "n_epochs": 100}
    types = [None, None]
    prefix = ".hdf5"
    loss = "binary_crossentropy"

    def __init__(self, params, weights):
        super().__init__(params, weights)
        self.model = self.load()

    def load(self):  
        # WHOLE CELL PREDICTION
        model = unet(pretrained_weights = self.weights,
                    input_size = (None,None,1))
        return model
    
    def getProbIm(self, im):
        im = equalize_adapthist(im)
        (nrow, ncol) = im.shape
        row_add = 16-nrow%16
        col_add = 16-ncol%16
        padded = np.pad(im, ((0, row_add), (0, col_add)))
        results = self.model(padded[np.newaxis,:,:,np.newaxis])
        res = np.array(results[0,:,:,0])
        print(res.max())
        return res[:nrow, :ncol]
    
    def threshold(self, im):
        """
        Binarize an image with a threshold given by the user, or if the threshold is None, calculate the better threshold with isodata
        Param:
            im: a numpy array image (numpy array)
            th: the value of the threshold (feature to select threshold was asked by the lab)
        Return:
            bi: threshold given by the user (numpy array)
        """
        th = float(self.params["Threshold"])
        im2 = np.array(im).copy()
        if th == None:
            th = skimage.filters.threshold_isodata(im2)
        bi = im2
        bi[bi > th] = 255
        bi[bi <= th] = 0
        return bi

    def getMask(self, pred):
        th = self.threshold(pred)
        seg  = segment(th,pred, min_distance = int(self.params["Minimum Cell Distance"]))
        return seg











import pandas as pd
import numpy as np
from yeastvision.track.cell import getLifeData, getCellData
from yeastvision.track.lineage import LineageConstruction, has_daughters, get_daughters, get_generations, get_gametes, compute_mating_lineage
from yeastvision.utils import normalize_im

class Population():
    def __init__(self, name, replicates):
        self.name = name
        self.replicates = []
    
    def getData(self, property, replicates = None):
        if replicates:
            toGet = [replicate for replicate in replicates if replicate in self.replicates]
        else:
            toGet = self.replicates

        data = None
        for replicate in self.replicates:
            newdata = replicate.get_heatmap_data(property, self.name)
            if data is not None:
                data = TimeSeriesData.combine_population_data(data, newdata)
            else:
                data = newdata
        return data
    




class PopulationReplicate():
    populations = {}
    def __init__(self, experiment_idx, population_name, ts):
        self.population = population_name
        self.id = experiment_idx
        self.name = self.population+"-"+str(id)
        self.ts = ts

        if self.population not in PopulationReplicate.populations:
            PopulationReplicate.populations[self.population] =  Population(self.population, [self])
        else:
            PopulationReplicate.populations[self.population].replicates.append(self.ts)


        

class TimeSeriesData():
    name: str
    obj_data: pd.core.frame.DataFrame
    obj_data_export: pd.core.frame.DataFrame
    life_data: pd.core.frame.DataFrame
    cell_data: pd.core.frame.DataFrame
    properties: list[str]
    populations:list[list[int]]
    population_names: list[str]
    label_vals: np.ndarray
    num_objs: int

    def __init__(self, idx, labels, channels = None, channel_names = None, cell_data = None, life_data = None):
        self.i = idx
        self.set_label_props(labels)
        self.populations = [self.label_vals-1]
        self.population_names = ["all"]
        
        
        self.channel_names = channel_names
        
        if cell_data is not None:
            self.cell_data = cell_data
        else:
            self.cell_data = getCellData(labels, channels, channel_names)
        
        if life_data is not None:
            self.life_data = life_data
        else:
            self.life_data = getLifeData(labels)
        
        self.update_props()

    def update_cell_data(self, labels, channels = None, channel_names = None):
        self.set_label_props(labels)
        self.life_data = getLifeData(labels)
        self.cell_data = getCellData(labels, channels, channel_names)
        self.channel_names = channel_names
        self.update_props()


    def update_props(self):
        self.properties = self.cell_data.columns.tolist()
        self.properties.remove("labels")
        print(self.properties)


    def set_label_props(self, labels):
        label_vals = np.unique(labels)
        self.label_vals = label_vals[label_vals>0]
        self.num_objs = len(self.label_vals)
    
    def get_population_data(self, column_name, population_name = "all", func = np.mean):
        population = self.populations[self.population_names.index(population_name)]
        df = np.array(self.cell_data.iloc[population][column_name].to_list(), dtype = float)
        return func(df, axis = 0)
    
    def get_heatmap_data(self, column_name, population_name):
        heatmap = normalize_im(np.array(self.cell_data[column_name].to_list(), dtype =float), clip =False)
        population = self.populations[self.population_names.index(population_name)]
        im = np.zeros_like(heatmap, dtype = float)
        im[im==0] = np.nan
        im[population, :] = heatmap[population, :]
        return im
    
    def add_population(self, cells, name):
        self.populations.append(list(np.array(cells)-1))
        self.names.append(name)
    
    def get_population_sample_size(self, population_name):
        return len(self.populations[self.population_names.index(population_name)])
    
    def get_cell_info(self):
        return pd.DataFrame.from_dict({"cell":self.label_vals} | self.life_data )

    def get(self, property, cell):
        data = np.array((self.cell_data[property][cell-1].copy()), dtype = float)
        print(data)
        return data

    def hasLineage(self):
        return False
    
    def hasMating(self):
        return False
    
    @staticmethod
    def combine_population_data(arr1, arr2):
        # Find the widths of the arrays
        width1 = arr1.shape[1]
        width2 = arr2.shape[1]

        # Calculate the difference in widths
        width_diff = abs(width1 - width2)

        # Pad the array with the shorter width with None values
        if width1 < width2:
            padding = [(0, 0), (0, width_diff)]
            arr1 = np.pad(arr1, padding, mode='constant', constant_values=None)
        elif width2 < width1:
            padding = [(0, 0), (0, width_diff)]
            arr2 = np.pad(arr2, padding, mode='constant', constant_values=None)

        # Concatenate the arrays vertically
        result = np.concatenate((arr1, arr2), axis=0)

        return result


class LineageData(TimeSeriesData):
    mothers:pd.core.frame.DataFrame
    daughters:pd.core.frame.DataFrame
    mating: pd.core.frame.DataFrame
    generations: np.ndarray
    
    _hasLineage:bool
    _hasMating:bool

    def __init__(self, idx, cells, buds = None, mating = None, cell_data = None, channels = None, channel_names = None, life_data = None):
        super().__init__(idx, cells, channels = channels, channel_names=channel_names, cell_data=cell_data, life_data=life_data)

        self._hasLineage = False
        if buds is not None:
            self.add_lineages(cells,buds)
        
        self._hasMating = False
        if mating is not None:
            self.add_mating(cells, mating)
    
    def add_lineages(self, cells, buds):
        self._hasLineage = True
        lineage = LineageConstruction(cells, buds, forwardskip=3)
        self.daughters, self.mothers = lineage.computeLineages()
        self.set_generations()
    
    def add_mating(self, cells, mating):
        self._hasMating = True
        self.mating = compute_mating_lineage(mating, cells)
        self.set_mating_data(mating)
        self.set_gamete_data()

    def get_initial_generation(self):
        life_df = pd.DataFrame.from_dict(self.life_data)
        potential_gen1 = life_df[life_df["birth"]==0].index.values
        return [val+1 for val in potential_gen1 if pd.isnull(self.mothers.iloc[val, self.mothers.columns.get_loc("mother")])]
    
    def set_generations(self):
        initial_gen = self.get_initial_generation()
        generations = get_generations(self.daughters, initial_gen)
        for i, generation in enumerate(generations):
            name = f"gen{i}"
            if name not in self.population_names:
                self.population_names.append(name)
                self.populations.append(np.array(generation, dtype=np.uint16)-1)
            else:
                idx = self.population_names.index(name)
                self.populations[idx] = (np.array(generation, dtype=np.uint16)-1)
        print(self.population_names)


    def set_mating_data(self, mating):
        mating_vals = np.unique(mating)
        mating_vals = mating_vals[mating_vals>0]
        self.add_population(mating_vals, "mating")
    
    def set_gamete_data(self):
        gametes = self.mating["gamete1"].values().to_list() + self.mating["gamete2"].values()
        gametes = [gamete for gamete in gametes if pd.notnull(gamete) and gamete not in gametes]
        self.add_population(gametes, "gametes")

    def get_cell_info(self):
        cell_info = {"cell":self.label_vals} | self.life_data 
        if self._hasLineage:
            cell_info = cell_info | self.mothers.to_dict(orient = "list")
        if self._hasMating:
            cell_info = cell_info | self.mating.to_dict(orient = "list")

        return pd.DataFrame.from_dict(cell_info)

    def hasLineages(self):
        return self._hasLineage
    
    def hasMating(self):
        return self._hasMating
    








import numpy as np
from skimage.measure import regionprops
from skimage.morphology import thin, skeletonize, opening, dilation, erosion
import statistics
from yeastvision.track.utils import get_bbox_coords
import statistics as stats
from yeastvision.track.cell import getCellData
import matplotlib.pyplot as plt
from tqdm import tqdm

def track_to_cell(objs, cells):
    tracked_mask = np.zeros_like(objs)
    im_idx = 0 
    for obj_mask, cell_mask in zip(objs, cells):
        for obj_num in np.unique(obj_mask):
            if obj_num > 0:
                vals, counts  = np.unique(cell_mask[obj_mask==obj_num], return_counts=True)

                if np.any(vals==0):
                    vals, counts = vals[1:], counts[1:]

                tracked_mask[im_idx][obj_mask==obj_num] = vals[np.argmax(counts)]
        im_idx += 1
    return tracked_mask

def extract_cell_frames(ims):
    found_cells = False
    for i, im in enumerate(ims):
        if not found_cells:
            found_cells = np.any(im>0)
        if found_cells:
            if np.all(im==0):
                break
    return i

def trackYeasts(ims):
    stop_i = extract_cell_frames(ims)
    ims_template = np.zeros_like(ims, dtype=np.uint16)
    ims = ims[:stop_i]
    print("--TRACKING: LOOP 1/4")
    IS1 = ims[0,:,:]
    IblankG = np.zeros(IS1.shape, dtype="uint16") # allocate
    masks = np.zeros((IS1.shape[0], IS1.shape[1], ims.shape[0])) # allocate
    masks[:,:,0]= IS1   

    for it0 in tqdm(range(1,ims.shape[0])):
        m0b = ims[it0]
        IS2 = m0b.copy()
        IS2C = IS2 
        IS1B = IS1 
        IS1B= (IS1 > 0.5).astype(np.uint16) # binarize # plt.imshow(IS1B)
        
        IS3 = np.multiply(IS1B,IS2) # plt.imshow(IS3)
        
        tr_cells=(np.unique(IS1.astype(np.int0)))
        tr_cells=tr_cells[1:len(tr_cells)]
        gap_cells=(np.unique(IblankG.astype(np.int0)))
        gap_cells=gap_cells[1:len(gap_cells)]
        
        # gap_cells=np.array((10,122)).astype(np.uint64) 
        # cells_tr =  cells_tr + gap_cells
        if gap_cells.size==0 :
            cells_tr = tr_cells;
        else:
        # cells_tr = np.array([tr_cells, gap_cells]) # !
        # cells_tr =  (tr_cells) + (gap_cells)
            cells_tr = np.concatenate((tr_cells,gap_cells),axis=0) # right concatenation
            cells_tr = np.sort(cells_tr)
            
        Iblank0=np.zeros(IS1.shape, dtype="uint16") # plt.imshow(Iblank0)
    
        for it1 in cells_tr: # ascending, default
            # it1=1
            IS5=(IS1==it1) # plt.imshow(IS5) # sum(sum(IS5))
            IS5=IS5.astype(np.uint16) # plt.imshow(IS5)
            IS56 = thin(IS5,1) # plt.imshow(IS56)  #  sum(sum(IS56))
            IS6A = np.multiply(IS56,IS3) # plt.imshow(IS6A)  #   sum(sum(IS6A))
        
            if sum(sum(IS5))==0 :
                IS5=(IblankG==it1)
                IS6A = np.multiply(IS56,IS2C)
                IblankG[IblankG==it1]=0
                # IblankG=np.where(IblankG==it1,IblankG,0)
            
            if sum(sum(IS6A))>0 :
                IS2ind=(statistics.mode(IS6A[np.nonzero(IS6A)])) # nonzero gives the indexes
                Iblank0[IS2==IS2ind]=it1
                IS3[IS3==IS2ind]=0;
                IS2C[IS2==IS2ind]=0;
                


        bl_cells = np.unique(Iblank0) # plt.imshow(Iblank0)
        bl_cells = bl_cells[1:len(bl_cells)]
        seg_gap = np.setdiff1d(tr_cells,bl_cells) 
        
                
        if seg_gap.size>0 :
            for itG in seg_gap:
                IblankG[IS1==itG]=itG;
        
        
        Iblank0B = np.copy(Iblank0)      # plt.imshow(Iblank0)
        Iblank0B[np.nonzero(Iblank0B)] = 1;  
        Iblank0B = (Iblank0B < 0.5).astype(np.uint16) # plt.imshow(Iblank0B) # invert Iblank0B
        
        ISB = np.multiply(IS2,Iblank0B) # plt.imshow(ISB)
        
        newcells=np.unique(ISB)
        newcells = newcells[1:len(newcells)]
        Iblank=Iblank0  # plt.imshow(Iblank)
        A=1;
        
        if newcells.size>0 :
            for it2 in newcells :
                Iblank[IS2==it2]=max(cells_tr)+A; 
                A=A+1;      
                
        masks[:,:,it0]=Iblank # plt.imshow(masks[:,:,it0-1])
        IS1=masks[:,:,it0] # plt.imshow(IS1)


    ccell2=np.unique(masks[:,:,len(masks[0][0])-1]);
    ccell2 = ccell2[1:len(ccell2)] # remove zero since its background
    Mask2 = np.zeros((masks.shape[0], masks.shape[1], len(masks[0][0])))

    print("TRACKING: LOOP 2/4")
    for itt4 in tqdm(range(0,len(masks[0][0]))) : # masks
        mask3=np.zeros((masks.shape[0], masks.shape[1]))
                
                
        for itt3 in range(0,len(ccell2)) : # % itt3=23
            pix2 = np.nonzero(masks[:,:,itt4]==ccell2[itt3])
            mask3[pix2]=itt3 + 1 # zero correction    
                
        Mask2[:,:,itt4] = mask3 

    no_obj=len(ccell2);
    all_ob=np.zeros((no_obj,len(Mask2[0][0])))
    print("TRACKING: LOOP 3/4")
    for ccell in tqdm(range(0,no_obj)): # ccell=0
            
            for itt in range(0,len(masks[0][0])): # itt = 0
                Mas=Mask2[:,:,itt]==ccell + 1 # % imagesc(Mask2)
                pix=sum(sum(Mas))
                all_ob[ccell,itt]=pix #  figure;imagesc(all_ob)

    cell_exists=np.zeros((1,len(all_ob[:,0]))).astype("uint16")
    print("TRACKING: LOOP 4/4")
    for itt2 in tqdm(range(0,len(all_ob[:,0]))) : # itt2 = 20
        etime = sum(all_ob[itt2,:]==0) #
        if etime == 0 :
            cell_exists[0,itt2]=1
        else:
            cell_exists[0,itt2]=np.max(np.nonzero(all_ob[itt2,:]==0)) + 1

    Mask2 = np.array([Mask2[:,:,i] for i in range(Mask2.shape[-1])], dtype = np.uint16)
    ims_template[:stop_i] = Mask2
    return ims_template



    




from scipy import ndimage as ndi
from skimage.feature import peak_local_max
from skimage.morphology import dilation
from skimage.segmentation import watershed
from skimage.filters import gaussian
from skimage.measure import label

import numpy as np


def segment(th, pred, min_distance=10, topology=None): 
    """
    Performs watershed segmentation on thresholded image. Seeds have to
    have minimal distance of min_distance. topology defines the watershed
    topology to be used, default is the negative distance transform. 
    Can either be an array with the same size af th, or a function that will
    be applied to the distance transform.
   , 
    After watershed, the borders found by watershed will be evaluated in terms
    of their predicted value. If the borders are highly predicted to be cells,
    the two cells are merged. 
    """
    dtr = ndi.morphology.distance_transform_edt(th)
    if topology is None:
        topology = -dtr
    elif callable(topology):
        topology = topology(dtr)

    indices = peak_local_max(-topology, min_distance = min_distance)
    m = np.zeros_like(th, dtype = bool)
    m[tuple(indices.T)] = True

    # Uncomment to start with cross for every pixel instead of single pixel
    m_lab = label(m) #comment this
    #m_dil = dilation(m)
    #m_lab = label(m_dil)
    wsh = watershed(topology, m_lab, mask=th, connectivity=2)
    merged = cell_merge(wsh, pred)
    return np.array(correct_artefacts(merged), dtype = np.uint16)
    
    
def correct_artefacts(wsh):
    """
    Sometimes artefacts arise with 3 or less pixels which are surrounded entirely
    by another cell. Those are removed here.
    """
    unique, count = np.unique(wsh, return_counts=True)
    to_remove = unique[count<=3]
    for rem in to_remove:
        rem_im = wsh==rem
        rem_cont = dilation(rem_im) & ~rem_im
        vals, val_counts = np.unique(wsh[rem_cont], return_counts=True)
        replace_val = vals[np.argmax(val_counts)]
        if replace_val != 0:
            wsh[rem_im] = int(replace_val)
    return wsh


def cell_merge(wsh, pred):
    """
    Procedure that merges cells if the border between them is predicted to be
    cell pixels.
    """
    wshshape=wsh.shape
    
    # masks for the original cells
    objs = np.zeros((wsh.max()+1,wshshape[0],wshshape[1]), dtype=bool)	
    
    # masks for dilated cells
    dil_objs = np.zeros((wsh.max()+1,wshshape[0],wshshape[1]), dtype=bool)
    
    # bounding box coordinates	
    obj_coords = np.zeros((wsh.max()+1,4))
    
    # cleaned watershed, output of function	
    wshclean = np.zeros((wshshape[0],wshshape[1]))
    
    # kernel to dilate objects
    kernel = np.ones((3,3), dtype=bool)	
    
    for obj1 in range(wsh.max()):
        # create masks and dilated masks for obj
        objs[obj1,:,:] = wsh==(obj1+1)	
        dil_objs[obj1,:,:] = dilation(objs[obj1,:,:], kernel)	
        
        # bounding box
        obj_coords[obj1,:] = get_bounding_box(dil_objs[obj1,:,:])
    
    objcounter = 0	# counter for new watershed objects
    
    for obj1 in range(wsh.max()):	
        dil1 = dil_objs[obj1,:,:]

        # check if mask has been deleted
        if np.sum(dil1) == 0:
            continue
        
        objcounter = objcounter + 1
        orig1 = objs[obj1,:,:]
        
        last_obj2_added_to_obj1 = -1

        for obj2 in range(obj1+1,wsh.max()):
            dil2 = dil_objs[obj2,:,:]
            
            # only check border if bounding box overlaps, and second mask 
            # is not yet deleted
            if (do_box_overlap(obj_coords[obj1,:], obj_coords[obj2,:])
                and np.sum(dil2) > 0):
                
                border = dil1 * dil2	
                border_pred = pred[border]
                
                # Border is too small to be considered
                if len(border_pred) < 16:	#SJR: Changed on 18.04.2021 from previously 32. Not sure why 32. I remember 8.
                    continue
                
                # Sum of top 25% of predicted border values
                q75 = np.quantile(border_pred, .75)
                top_border_pred = border_pred[border_pred >= q75]
                top_border_height = top_border_pred.sum()
                top_border_area = len(top_border_pred)
                
                # merge cells
                if top_border_height / top_border_area > .99:
                    orig1 = np.logical_or(orig1, objs[obj2,:,:])
                    dil1 = np.logical_or(dil1, dil2)
                    dil_objs[obj1,:,:] = dil1
                    dil_objs[obj2,:,:] = np.zeros((wshshape[0], wshshape[1]))
                    obj_coords[obj1,:] = get_bounding_box(dil_objs[obj1,:,:])
#                    obj_coords[obj2,:] = get_bounding_box(dil_objs[obj2,:,:])
                    last_obj2_added_to_obj1 = obj2
        
        # the last object that obj1 was merged with should be equal to obj1 so that additional cells could be merged with it
        if last_obj2_added_to_obj1 > -1:
            obj2 = last_obj2_added_to_obj1
            dil_objs[obj2,:,:] = dil1
            objs[obj2,:,:] = orig1
            obj_coords[obj2,:] = get_bounding_box(dil_objs[obj2,:,:])
            
                    
        wshclean = (1-orig1)*wshclean + orig1*objcounter
        
    # resort wshclean
    u = np.unique(wshclean)[1:]	#ignore background
    for obj1 in range(len(u)):
        wshclean[wshclean==u[obj1]] = obj1 + 1
            
    return wshclean


def do_box_overlap(coord1, coord2):
    """Checks if boxes, determined by their coordinates, overlap. Safety
    margin of 2 pixels"""
    return (
    (coord1[0] - 2 < coord2[0] and coord1[1] + 2 > coord2[0]
        or coord2[0] - 2 < coord1[0] and coord2[1] + 2 > coord1[0]) 
    and (coord1[2] - 2 < coord2[2] and coord1[3] + 2 > coord2[2]
        or coord2[2] - 2 < coord1[2] and coord2[3] + 2 > coord1[2]))

    
def get_bounding_box(im):
    """Returns bounding box of object in boolean image"""
    coords = np.where(im)
    
    return np.array([np.min(coords[0]), np.max(coords[0]), 
                     np.min(coords[1]), np.max(coords[1])])


from setuptools import setup

requires = [
"numpy==1.24",
"cellpose==2.1.0",
"matplotlib",
"munkres",
"opencv_python_headless",
"pandas==1.5.0",
"patchify==0.2.3",
"PyQt5==5.15.9",
"pyqtgraph==0.13.0",
"scikit_image",
"scikit_learn",
"scipy",
"tensorflow==2.10.0",
"torch==1.12.0",
"tqdm",
"trackpy",
"torchvision==0.13.0",
"memory-profiler"]

try:
    import torch
    a = torch.ones(2, 3)
    major_version, minor_version, _ = torch.__version__.split(".")
    if major_version == "2" or int(minor_version) >= 12:
        requires.remove("torch==1.12.0")
        requires.remove("torchvision==0.13.0")
except:
    pass

with open("README.md", "r") as fh:
    long_description = fh.read()

packages = ["yeastvision", "yeastvision.plot", "yeastvision.track", "yeastvision.models", "yeastvision.ims", "yeastvision.ims.rife_model",
            "yeastvision.ims.rife_model.pytorch_msssim", "yeastvision.parts", "yeastvision.flou", "yeastvision.disk", 
            "yeastvision.models.artilife", "yeastvision.models.artilife.budSeg", "yeastvision.models.artiBud",
            "yeastvision.models.matSeg", "yeastvision.models.tetradSeg", "yeastvision.models.budNET", 
            "yeastvision.models.vacNET", "yeastvision.models.YeaZ"]


setup(
    name = "yeastvision",
    version = "0.1.16",
    description = "Deep learning-enabled image analysis of the yeast full life cycle",
    author = "Berk Yalcinkaya",
    url = "https://github.com/berkyalcinkaya/yeastvision",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author_email="berkyalcinkaya55@gmail.com",
    license = "BSD",
    packages = packages,
    install_requires = requires,
    entry_points = {
        'console_scripts': [
          'yeastvision = yeastvision.__main__:main',
          'install-weights = yeastvision.install_weights:do_install']
       }
)



from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Conv2D, MaxPooling2D, Dropout, 
                                     concatenate, UpSampling2D)
from tensorflow.keras.optimizers import Adam


def unet(pretrained_weights = None,input_size = (256,256,1)):
    inputs = Input(input_size)
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(inputs)
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool1)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool2)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool3)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv4)
    drop4 = Dropout(0.5)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool4)
    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv5)
    drop5 = Dropout(0.5)(conv5)

    up6 = Conv2D(512, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(drop5))
    merge6 = concatenate([drop4,up6], axis = 3)
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge6)
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv6)

    up7 = Conv2D(256, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv6))
    merge7 = concatenate([conv3,up7], axis = 3)
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge7)
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv7)

    up8 = Conv2D(128, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv7))
    merge8 = concatenate([conv2,up8], axis = 3)
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge8)
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv8)

    up9 = Conv2D(64, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv8))
    merge9 = concatenate([conv1,up9], axis = 3)
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge9)
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
    conv9 = Conv2D(2, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
    conv10 = Conv2D(1, 1, activation = 'sigmoid')(conv9)

    model = Model(inputs = inputs, outputs = conv10)

    model.compile(optimizer = Adam(learning_rate = 1e-4), loss = 'binary_crossentropy', metrics = ['accuracy'])

    if(pretrained_weights):
        model.load_weights(pretrained_weights)

    return model


import cv2
from skimage.io import imread
import matplotlib.pyplot as plt
from torch.nn import functional as F
import torch
import numpy as np
from yeastvision.ims.rife_model.pytorch_msssim import ssim_matlab
from queue import Queue, Empty
from skimage.util import img_as_ubyte
from .rife_model.RIFE import Model
import os
from yeastvision.ims import rife_model

RIFE_DIR = rife_model.__path__[0]

def interpolate(ims: np.ndarray, exp: int)->np.ndarray:
    scale = 1
    
    def make_inference(I0, I1, n):
        middle = model.inference(I0, I1, scale)
        if n == 1:
            return [middle]
        first_half = make_inference(I0, middle, n=n//2)
        second_half = make_inference(middle, I1, n=n//2)
        if n%2:
            return [*first_half, middle, *second_half]
        else:
            return [*first_half, *second_half]
    
    def pad_image(img):
        tmp = 32
        ph = ((h - 1) // tmp + 1) * tmp
        pw = ((w - 1) // tmp + 1) * tmp
        padding = (0, pw - w, 0, ph - h)
        return F.pad(img, padding)
    
    def reformat(im):
        im = cv2.normalize(im, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
        return  cv2.merge((im, im, im))
    
    def make_read_buffer(ims):
        read_buffer = Queue(maxsize=500)
        for im in ims:
            read_buffer.put(im)
        return read_buffer

    ims = ims = [reformat(im) for im in ims]
    lastframe = ims[0]
    h,w,_ = lastframe.shape
    ims = ims[1:]

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    torch.set_grad_enabled(False)
    if torch.cuda.is_available():
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True
    
    model = Model()
    model.load_model(RIFE_DIR, -1)
    model.eval()
    model.device()
    

    write_buffer = []
    read_buffer = make_read_buffer(ims)

    I1 = torch.from_numpy(np.transpose(lastframe, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
    I1 = pad_image(I1)
    temp = None # save lastframe when processing static frame
    i = 0
    while True:
        if temp is not None:
            frame = temp
            temp = None
        else:
            try:
                frame = read_buffer.get(False)
            except Empty:
                break
        I0 = I1
        I1 = torch.from_numpy(np.transpose(frame, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
        I1 = pad_image(I1)
        I0_small = F.interpolate(I0, (32, 32), mode='bilinear', align_corners=False)
        I1_small = F.interpolate(I1, (32, 32), mode='bilinear', align_corners=False)
        ssim = ssim_matlab(I0_small[:, :3], I1_small[:, :3])

        break_flag = False
        if ssim > 0.996:    
            try:    
                frame = read_buffer.get(False) # read a new frame
                temp = frame
            except Empty:
                break_flag = True
                frame = lastframe

            I1 = torch.from_numpy(np.transpose(frame, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
            I1 = pad_image(I1)
            I1 = model.inference(I0, I1, scale)
            I1_small = F.interpolate(I1, (32, 32), mode='bilinear', align_corners=False)
            ssim = ssim_matlab(I0_small[:, :3], I1_small[:, :3])
            frame = (I1[0] * 255).byte().cpu().numpy().transpose(1, 2, 0)[:h, :w]
        
        if ssim < 0.2:
            output = []
            for i in range((2 ** exp) - 1):
                output.append(I0)

        else:
            output = make_inference(I0, I1, 2**exp-1) if exp else []

        write_buffer.append(lastframe)
        for mid in output:
            mid = (((mid[0] * 255.).byte().cpu().numpy().transpose(1, 2, 0)))
            write_buffer.append(mid[:h, :w])

        lastframe = frame
        if break_flag:
            break
        print(i)
        i+=1
    del model
    return np.array([im[:,:,0] for im in write_buffer], dtype = np.uint8)


import shutil
import os 
from os.path import join
import glob
import numpy as np
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtCore import Qt
from yeastvision.models.utils import MODEL_DIR


def loadCustomModel(parent):
    filter = "Archive File (*);; HDF File (*.h5 *.hdf5)"
    currDir = parent.getCurrImDir() if parent.imLoaded else os.getcwd()
    path,_ = QFileDialog.getOpenFileName(
        parent = parent,
        caption = f'''Select Custom Weights to Load. The name of the model architecture should be included in the filename.Current models are {parent.modelTypes}''',
        directory = currDir,
        filter = filter
    )

    if not path:
        return

    customModelFile = os.path.split(path)[1]
    customModelName = customModelFile.split(".")[0]
    validName = False
    for modelType in parent.modelTypes:
        if modelType.lower() in customModelName.lower():
            validName = True
            break
    if not validName:
        parent.showError("Custom Model Must Have Model Type in File Name")
        return
    dst = join(MODEL_DIR, modelType, customModelFile)
    shutil.copy2(path,dst)
    parent.updateModelNames()

